import { h, s } from './h.js'

const COLORS = ['#e22', '#e73', '#fc3', '#ad4', '#4d9', '#3be', '#45d', '#c3e']
const PADDING = 15

const _svg = s('svg', { viewBox: '0 0 0 0', width: 0, height: 0 })

h(document.body, ['div', { $style: { width: 0, height: 0 } }, _svg])

const toFixed2 = (/** @type {number} */ value) => value.toFixed(2)

/**
 * A helper for creating a chart based on a table (conf. <https://jackens.github.io/nnn/chartable/>).
 *
 * Options:
 * - `table`: `HTMLTableElement` to extract data, data series labels and X axis labels
 * - `headerColumn`: flag indicating that `table` has a header column with X axis labels
 * - `xGap`: X axis spacing
 * - `xLabelsMinHeight`: minimal height of X axis labels
 * - `xLabelsRotate`: flag to rotate X axis labels
 * - `xReverse`: flag to reverse all data series
 * - `yGap`: Y axis spacing
 * - `yLabelsLeftMinWidth`: minimal width of Y axis left labels
 * - `yLabelsRightMinWidth`: minimal width of Y axis right labels
 * - `yMax`: number of Y axis lines
 * - `zLabelsMinWidth`: minimal width of data series labels
 * - `zyMappings`: mappings per data series
 *
 * @param {{
 *   table: HTMLTableElement;
 *   headerColumn?: boolean;
 *   xGap?: number;
 *   xLabelsMinHeight?: number;
 *   xLabelsRotate?: boolean;
 *   xReverse?: boolean;
 *   yGap?: number;
 *   yLabelsLeftMinWidth?: number;
 *   yLabelsRightMinWidth?: number;
 *   yMax?: number;
 *   zLabelsMinWidth?: number;
 *   zyMappings?: [(label: string) => number, (value: number) => string][];
 * }} options
 */
export const chartable = ({
  table,
  headerColumn = false,
  xGap = 70,
  xLabelsMinHeight = 0,
  xLabelsRotate = false,
  xReverse = false,
  yGap = 30,
  yLabelsLeftMinWidth = 100,
  yLabelsRightMinWidth = 100,
  yMax = 10,
  zLabelsMinWidth = 0,
  zyMappings = []
}) => {
  const /** @type {number[][]} */ zxValues = []
  const /** @type {string[]}   */ xLabels = []
  const /** @type {string[]}   */ zLabels = []

  table.querySelectorAll('tr').forEach((row, r) =>
    // @ts-expect-error
    row.querySelectorAll('td,th').forEach((/** @type {HTMLTableCellElement} */ col, c) => {
      const x = r - 1
      const z = c - +headerColumn
      const value = col.innerText

      if (x >= 0 && z >= 0) {
        zxValues[z] = zxValues[z] ?? []
        zxValues[z][x] = (zyMappings?.[z]?.[0] ?? parseFloat)(value)
      } else if (x >= 0 && z < 0) {
        xLabels[x] = value
      } else if (x < 0 && z >= 0) {
        zLabels[z] = value
      }
    }))

  if (xReverse) {
    xLabels.reverse()
    zxValues.forEach(xValues => xValues.reverse())
  }

  const xMax = zxValues[0].length
  const xPx = Array.from({ length: xMax }, (_, x) => x * xGap)
  const yPx = Array.from({ length: yMax }, (_, y) => y * yGap)

  const width = xGap * (xMax - 1)
  const height = yGap * (yMax - 1)

  const gGraph = s('g',
    ...yPx.map(px => ['line', { x1: 0, x2: width, y1: px, y2: px, stroke: '#8888' }]),
    ...xPx.map(px => ['line', { x1: px, x2: px, y1: 0, y2: height, stroke: '#8888' }]))
  const gXLabels = s('g')
  const gYLabelsL = zxValues.map(_ => s('g'))
  const gYLabelsR = zxValues.map(_ => s('g'))
  const gZColors = s('g')
  const /** @type {SVGTextElement[]} */ gZLabels = []

  const onclick = (/** @type {number} */ z) => {
    gYLabelsL.forEach((g, z2) => s(g, { $style: { display: z2 === z ? 'block' : 'none' } }))
    gYLabelsR.forEach((g, z2) => s(g, { $style: { display: z2 === z ? 'block' : 'none' } }))
    gZLabels.forEach((text, z2) => s(text, { $style: { fontWeight: z2 === z ? 'bold' : 'normal', cursor: 'pointer' } }))
  }

  zxValues.forEach((values, z) => {
    const fill = COLORS[z % COLORS.length]
    const minValue = Math.min(...values.filter(value => !isNaN(value)))
    const maxValue = Math.max(...values.filter(value => !isNaN(value)))

    yPx.forEach((yp, y) => {
      const value = (zyMappings?.[z]?.[1] ?? toFixed2)(maxValue - (maxValue - minValue) / yMax * (y + 1))
      const textL = s('text', { x: 0, y: yp, 'text-anchor': 'end', 'alignment-baseline': 'middle' }, value)
      const textR = s('text', { x: 0, y: yp, 'text-anchor': 'start', 'alignment-baseline': 'middle' }, value)

      s(_svg, textL, textR)

      yLabelsLeftMinWidth = Math.max(yLabelsLeftMinWidth, textL.getBBox().width)
      yLabelsRightMinWidth = Math.max(yLabelsRightMinWidth, textR.getBBox().width)

      s(gYLabelsL[z], textL)
      s(gYLabelsR[z], textR)
    })

    const text = s('text', {
      x: 0,
      y: yGap * z,
      'text-anchor': 'start',
      'alignment-baseline': 'middle',
      $onclick: () => onclick(z),
      $style: { fontWeight: 'bold' }
    },
    zLabels[z])

    s(_svg, text)

    zLabelsMinWidth = Math.max(zLabelsMinWidth, text.getBBox().width)

    gZLabels[z] = text

    s(gZColors, ['circle', { cx: 0, cy: yGap * z, r: 5, fill }])

    const valuesPx = values.map(value => isNaN(value) ? value : height * (maxValue - value) / (maxValue - minValue))

    xPx.forEach((px, x) => {
      if (!isNaN(valuesPx[x])) {
        s(gGraph, ['g', ['circle', { cx: px, cy: valuesPx[x], r: 5, fill }]])

        if (!isNaN(valuesPx[x - 1])) {
          s(gGraph, ['line', { x1: px, x2: xPx[x - 1], y1: valuesPx[x], y2: valuesPx[x - 1], stroke: fill }])
        }
      }
    })
  })

  if (xLabels.length > 0) {
    xPx.forEach((xp, x) => {
      const text = s('text',
        xLabelsRotate
          ? { x: 0, y: -xp, 'text-anchor': 'start', 'alignment-baseline': 'middle', transform: 'rotate(90)' }
          : { x: xp, y: PADDING, 'text-anchor': 'middle', 'alignment-baseline': 'middle' },
        xLabels[x])

      s(_svg, text)

      xLabelsMinHeight = Math.max(xLabelsMinHeight, xLabelsRotate ? text.getBBox().width : text.getBBox().height)

      s(gXLabels, text)
    })
  }

  const svgW = width + yLabelsLeftMinWidth + yLabelsRightMinWidth + zLabelsMinWidth + 4 * PADDING
  const svgH = height + xLabelsMinHeight + (xLabels.length > 0 ? 3 : 2) * PADDING

  s(gGraph, { transform: `translate(${yLabelsLeftMinWidth + PADDING} ${PADDING})` })

  gYLabelsL.forEach(g => s(g, { transform: `translate(${yLabelsLeftMinWidth} ${PADDING})` }))
  gYLabelsR.forEach(g => s(g, { transform: `translate(${width + yLabelsLeftMinWidth + 2 * PADDING} ${PADDING})` }))

  s(gZColors, { transform: `translate(${width + yLabelsLeftMinWidth + yLabelsRightMinWidth + 3 * PADDING} ${PADDING})` })
  s(gXLabels, { transform: `translate(${yLabelsLeftMinWidth + PADDING} ${height + 2 * PADDING})` })

  onclick(0)

  return s('svg',
    { viewBox: `0 0 ${svgW} ${svgH}`, width: `${svgW}px`, height: `${svgH}px`, class: 'chartable' },
    gGraph,
    gXLabels,
    ...gYLabelsL,
    ...gYLabelsR,
    gZColors,
    ['g', { transform: `translate(${width + yLabelsLeftMinWidth + yLabelsRightMinWidth + 4 * PADDING} ${PADDING})` }, ...gZLabels])
}

export const tests = {}
